
;+

function trace_mdi_align,tindex,tdata,mindex,mdata, $
                         trace_mask=rmaskin,mdi_mask=tmaskin, $
                         tmdi2traceguess=tmdi2traceguess, $
                         xvr=xvr,yvr=yvr,xvi=xvi,yvi=yvi, $
                         quiet=quiet,mdi2trace=mdi2trace, $
                         new_mdata=inew,new_tdata=rimage, $
                         _EXTRA=extra

;NAME:
;     TRACE_MDI_ALIGN
;PURPOSE:
;     Use an MDI white light image to fix the pointing in a TRACE
;     white light image by assuming that the MDI pointing is correct.
;CATEGORY:
;CALLING SEQUENCE:
;     new_trace_index = trace_mdi_align(tindex,tdata,mindex,mdata)
;INPUTS:
;     tindex  = TRACE index (a structure)
;     tdata   = TRACE image
;     mindex  = MDI index (a structure)
;     mdata   = MDI image
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS:
;     trace_mask = Index giving points in the TRACE image to neglect in the
;                  alignment.
;     mdi_mask = Index giving points in the MDI image to neglect in the
;                alignment.
;     mdi2trace = the final transformation from the MDI image to the TRACE
;                 image (output).  You probably don't really want this.
;     new_tdata = the piece of the TRACE image used to do the alignment
;                 (output).
;     new_mdata = the piece of the MDI images transformed onto new_tdata
;                 (output).
;     /quiet = work quietly
;
;     INPUT PARAMETERS FOR THE INITIAL GUESS (NOT REQUIRED):
;
;     If these are passed as input you will not have to click on points in
;     the two images which may be useful for automated alignments etc.,
;     however, if you use them, they must all be passed in and they must
;     all be consistent.  I.e., the transformation must apply to the subimages
;     defined by xvr,yvr,xvi,yvi.  If they are not consistent, the answer you 
;     get will be incorrect.  Garbage in, garbage out.  Be careful with these,
;     getting them right can be a bit tricky!
;
;     tmdi2traceguess = Tranform vector (from caltrans.pro) giving the initial
;                       guess for the alignment.  If this is not specified,
;                       the user will be prompted to click on corresponding
;                       points in the TRACE and MDI images in order to get
;                       the initial guess.  If this is passed, xvr,yvr,xvi,yvi
;                       must also be passed in.  If undefined, it will be
;                       set on output to the guess from setpts_roi.
;     xvr,yvr = 2 element arrays giving LL a UR corners of the TRACE subimage
;               to use.  Ignored unless tmdi2traceguess is also passed.
;               If undefined, it will be set on output to the guess from 
;               setpts_roi.
;     xvi,yvi = 2 element arrays giving LL a UR corners of the MDI subimage
;               to use.  Ignored unless tmdi2traceguess is also passed.
;               If undefined, it will be set on output to the guess from 
;               setpts_roi.
;OUTPUTS:
;     new_trace_index = trace index with the pointing (i.e. .xcen and .ycen)
;                       corrected.  .xcen and .ycen are the only things
;                       changed in the new index.
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     VERY SLOW
;
;     Obviously the TRACE and MDI images should be close in time to avoid
;     evolution in the AR (but solar rotation and foreshortening are 
;     accounted for).
;
;     MDI pointing is assumed to be correct.
;
;     I doubt the result is better than 0.5 arcsec and may not be that good.
;PROCEDURE:
;     The user first gives an initial guess for the transformation between
;     TRACE and MDI using setpts_roi and then the guess is refined using 
;     auto_align_image.   The transformation is then used to set the pointing
;     of TRACE assuming that the MDI pointing is correct.
;EXAMPLE:
;     ; First get the MDI data
;     mdata  = readfits('Some_MDI_FITS_File',smdihead)
;     mindex = fitshead2struct(smdihead)
;     ; Second get the TRACE data
;     trace_cat,some_start_time,some_end_time,cat
;     ss = where(cat.wave_len EQ 'WL')
;     trace_cat2data,cat[ss[0]],i,d
;     trace_prep,i,d,tindex,tdata,/wave2point
;     ; Third align the MDI and TRACE data
;     tindex_new = trace_mdi_align(tindex,tdata,mindex,mdata)
;MODIFICATION HISTORY:
;     T. Metcalf 2001-Oct-17
;     T. Metcalf 2001-Nov-14 Added a warning for a large coordinate shift.
;-

if strtrim(strupcase(tindex.wave_len),2) NE 'WL' then $
   message,'The TRACE image is not white light'

nxmdi = n_elements(mdata[*,0])
nymdi = n_elements(mdata[0,*])

bad = where(mdata LT mindex.d_min0 OR mdata GT mindex.d_max0,nbad)
if nbad GT 0 then mdata[bad] = 0

mdidate = anytim(mindex.date_obs,/yoh)

; Get the pixel size of MDI image

mdipix = mindex.im_scale
; This corrects the known problem with the MDI plate scale.  It might be
; necessary to update the value periodically???
if abs(mdipix-1.9778400) LT 0.00001 then begin
   mdipix = 1.982
   message,/info,'Correcting MDI plate scale of 1.97784 to '+string(mdipix)
endif
mdipix_earth = mdipix*mindex.EARTH_R0/mindex.OBS_R0

nxtrace = tindex.naxis1
nytrace = tindex.naxis2

if n_elements(tmdi2traceguess) LE 0 then begin

   ; Get the initial guess if one is not passed in

   message,/info,'Select the regions of the TRACE and MDI images to ' + $
                 'analyze.  Then click on corresponding points in the two ' + $
                 'images to set the initial guess for the transformation ' + $
                 'between them.'

   transtype = {type:'s', $
                ref_x:tindex.cdelt1, $
                ref_y:tindex.cdelt2, $
                pix_x:mdipix_earth, $
                pix_y:mdipix_earth, $
                transpose:0,phi:0.0}

   pp = setpts_roi(mdata,tdata,key=transtype, $
                   xvi=xvi,yvi=yvi,xvr=xvr,yvr=yvr)

   ; Get back to the restricted FOV since we'll use that in the 
   ; alignment

   pp(0,0,*) = pp(0,0,*) -  xvr(0)  ; x-coord of ref image
   pp(1,0,*) = pp(1,0,*) -  yvr(0)  ; y-coord of ref image
   pp(0,1,*) = pp(0,1,*) -  xvi(0)  ; x-coord of transformed image
   pp(1,1,*) = pp(1,1,*) -  yvi(0)  ; y-coord of transformed image

   tmdi2traceguess = caltrans(pp,transtype=transtype)
   if NOT keyword_set(quiet) then begin
      print,'Restricted initial guess for transformation:'
      pq2rss,tmdi2traceguess(*,0),tmdi2traceguess(*,1), $
         erot,exscl,eyscl,exshft,eyshft,enrss
      print
   endif
endif

rimage = tdata[xvr[0]:xvr[1],yvr[0]:yvr[1]]
timage = mdata[xvi[0]:xvi[1],yvi[0]:yvi[1]]

if n_elements(rmaskin) GT 0 then begin
   rmask = tdata & rmask[*] = 0.0
   rmask[rmaskin] = 1.0
   rmask = rmask[xvr[0]:xvr[1],yvr[0]:yvr[1]]
   rmask = where(rmask)
endif

if  n_elements(tmaskin) GT 0 then begin
   tmask = mdata & tmask[*] = 0.0
   tmask[tmaskin] = 1.0
   tmask = tmask[xvi[0]:xvi[1],yvi[0]:yvi[1]]
   tmask = where(tmask)
endif

; Use a 2nd degree transformation to allow for evolution and foreshortening
; due to solar rotation.

pin = tmdi2traceguess(*,0)
qin = tmdi2traceguess(*,1)
p9 = fltarr(3,3) & p9[0,0] = reform(pin,2,2) & p9 = reform(p9,9)
q9 = fltarr(3,3) & q9[0,0] = reform(qin,2,2) & q9 = reform(q9,9)

inew=auto_align_images(timage,rimage,p9,q9,pout,qout,missing=0.0, $
                       tmask=tmask,rmask=rmask,unrestricted=1, $
                       quiet=quiet,_EXTRA=extra)

mdi2trace=[[pout],[qout]]

; Get Lat/Lon of MDI data

tracedate = tindex.date_obs
tracedate = anytim(tracedate,/yoh)

ddays = int2secarr(anytim2ints(mdidate),anytim2ints(tracedate))/86400.
print,'ddays:',ddays

; get_latlon changes the pixs input to a 2-elements array so protect mdipix
jmdipix = mdipix  

latlon = get_latlon(nx=xvi[1]-xvi[0]+1L, $
                    ny=yvi[1]-yvi[0]+1L, $
                    x0=(xvi[0]+xvi[1])/2.0-mindex.x0, $
                    y0=(yvi[0]+yvi[1])/2.0-mindex.y0, $
                    b0=mindex.obs_b0*!dtor, $
                    pang=mindex.solar_p*!dtor, $  ; should be zero
                    pixs=jmdipix, $
                    rad=mindex.obs_r0)

; convert center of TRACE restricted FOV to MDI coordinates
 
xtrace = (xvr[1]-xvr[0])/2.0
ytrace = (yvr[1]-yvr[0])/2.0
pq2xy,xtrace,ytrace,mdi2trace[*,0],mdi2trace[*,1],xtrace_mdi,ytrace_mdi

; What is the lat/cmd of this MDI point?  Remove solar rotation
; here since that piece of the transformation is not a pointing shift.

lat = bilinear(latlon.latitude,xtrace_mdi,ytrace_mdi)
cmd = bilinear(latlon.cmd,xtrace_mdi,ytrace_mdi)-diff_rot(ddays,lat,/allen)

xy = lonlat2xy([cmd,lat],tracedate) ; x,y with solar rotation removed

new_tindex = tindex
new_tindex.xcen = ((nxtrace-1.)/2.-(xtrace+xvr[0]))*tindex.cdelt1 + xy[0]
new_tindex.ycen = ((nytrace-1.)/2.-(ytrace+yvr[0]))*tindex.cdelt2 + xy[1]

if NOT keyword_set(quiet) then begin
   print,'Old center: ',tindex.xcen,tindex.ycen
   print,'New center: ',new_tindex.xcen,new_tindex.ycen
   print,'TRACE off by ',$
         sqrt((tindex.xcen-new_tindex.xcen)^2 + $
              (tindex.ycen-new_tindex.ycen)^2),' arcsec'
endif

if sqrt((tindex.xcen-new_tindex.xcen)^2 + $
        (tindex.ycen-new_tindex.ycen)^2) GT 20. then $
   message,/info,'WARNING: The pointing change is large.  ' + $
                 'It is possible that the routine converged to an '+$
                 'incorrect solution.  Try re-running and pass the '+$
                 'keywords "/powell,scale=0.1".'

return,new_tindex

end
